DEFAULT_GET_WORDS_MATCH_REGEX = r'/\w+/g'
DEFAULT_GET_WORDS_JOIN_WITH = r'"\n"'

DEFAULT_GET_TEXT_DELIMITER_REGEX = r'/\n|\r|\t/g'
DEFAULT_GET_TEXT_REPLACE_WITH = r'" "'

DEFAULT_GET_HTML_DELIMITER_REGEX = r'/\n|\r|\t/g'
DEFAULT_GET_HTML_REPLACE_WITH = r'" "'
