from brotab.const import DEFAULT_GET_HTML_DELIMITER_REGEX
from brotab.const import DEFAULT_GET_HTML_REPLACE_WITH
from brotab.const import DEFAULT_GET_TEXT_DELIMITER_REGEX
from brotab.const import DEFAULT_GET_TEXT_REPLACE_WITH
from brotab.const import DEFAULT_GET_WORDS_JOIN_WITH
from brotab.const import DEFAULT_GET_WORDS_MATCH_REGEX
from brotab.utils import encode_query

DEFAULT_TRANSPORT_TIMEOUT = 60.0
DEFAULT_SHUTDOWN_POLL_INTERVAL = 1.0
DEFAULT_HTTP_IFACE = '127.0.0.1'
DEFAULT_MIN_HTTP_PORT = 4625
DEFAULT_MAX_HTTP_PORT = DEFAULT_MIN_HTTP_PORT + 10
DEFAULT_GET_WORDS_MATCH_REGEX = encode_query(DEFAULT_GET_WORDS_MATCH_REGEX)
DEFAULT_GET_WORDS_JOIN_WITH = encode_query(DEFAULT_GET_WORDS_JOIN_WITH)
DEFAULT_GET_TEXT_DELIMITER_REGEX = encode_query(DEFAULT_GET_TEXT_DELIMITER_REGEX)
DEFAULT_GET_TEXT_REPLACE_WITH = encode_query(DEFAULT_GET_TEXT_REPLACE_WITH)
DEFAULT_GET_HTML_DELIMITER_REGEX = encode_query(DEFAULT_GET_HTML_DELIMITER_REGEX)
DEFAULT_GET_HTML_REPLACE_WITH = encode_query(DEFAULT_GET_HTML_REPLACE_WITH)
